/*
 *   display.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "resource.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"
#include "kml.h"

#define DISPLAY_FREQ	16					// display update 1/frequency (1/64) in ms

#define B 0x00000000						// black
#define W 0x00FFFFFF						// white
#define I 0xFFFFFFFF						// ignore

// Sacajawea
#define PIX_PER_COL   7						// pixel per column
#define COL_PER_CHAR  5						// column per character

// Lewis
#define LCD_ROW_H		140					// max. pixel per line (131 or 137)
#define LCD_ROW_V32		 32					// max. pixel per line 32
#define LCD_ROW_V16		 16					// max. pixel per line 16

#define WIDTHBYTES(bits) (((bits) + 31) / 32 * 4)

#define ROP_PSDPxax 0x00B8074A				// ternary ROP
#define ROP_PDSPxax 0x00D80745

// convert color from RGBQUAD to COLORREF
#define RGBtoCOLORREF(c) ((((c) & 0xFF0000) >> 16) \
						| (((c) & 0xFF)     << 16) \
						|  ((c) & 0xFF00))

UINT    nFullScreen = 0;
UINT    nVertical = 0;
UINT    nBackgroundX = 0;
UINT    nBackgroundY = 0;
UINT    nBackgroundW = 0;
UINT    nBackgroundH = 0;
UINT    nLcdX = 0;
UINT    nLcdY = 0;

UINT    nLcdZoom = 1;						// display for Lewis
UINT    nLcdyZoom = (UINT) -1;				// not initialized

UINT	nLcdXOff = 0;						// display for Sacajawea
UINT	nLcdYOff = 0;
UINT	nLcdXPixel = 1;
UINT	nLcdYPixel = 1;
UINT	nLcdDistance = 0;

HDC     hLcdDC = NULL;
HDC     hAnnunDC = NULL;					// annunciator DC for Sacajawea
HDC     hMainDC = NULL;

static LPBYTE  pbyLcd;

static HBITMAP hLcdBitmap;
static HBITMAP hMainBitmap;

// Lewis
static UINT    nLcdxZoom;
static INT     nLcdWidth,nRot180;

// Sacajawea
static HDC     hSelAnnunDC;					// current annunciator DC
static HBITMAP hAnnunBitmap;

static HDC     hBmpBkDC;					// current background
static HBITMAP hBmpBk;

static HDC     hMaskDC;						// mask bitmap
static HBITMAP hMaskBitmap;
static LPBYTE  pbyMask;

static HBRUSH  hBrush;						// current segment drawing brush

static UINT nLcdXSize = 0;					// display size
static UINT nLcdYSize = 0;

static HANDLE hCThreadLcd = NULL;
static HANDLE hEventLcd;

static DWORD dwKMLColor[64] =				// color table loaded by KML script
{
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
};

static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	RGBQUAD bmiColors[2];
} bmiLcd =
{
	{ sizeof(BITMAPINFOHEADER),0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,ARRAYSIZEOF(bmiLcd.bmiColors),0 }
};

VOID (*UpdateContrast)(VOID) = NULL;

static VOID UpdateContrastSaca(VOID);
static VOID UpdateContrastLewis(VOID);

static VOID (*UpdateDisplayMem)(VOID) = NULL;

static VOID UpdateDisplayClamshellH(VOID);
static VOID UpdateDisplayPioneerH(VOID);
static VOID UpdateDisplayClamshellV(VOID);
static VOID UpdateDisplayPioneerV(VOID);
static VOID UpdateDisplaySacajawea(VOID);

//
// calculate size of LCD bitmap from KML script settings
//
static __inline VOID GetDimLcdBitmap(VOID)
{
	UINT nX,nY;
	INT  i;

	// get dimension of dot matrix part
	nLcdXSize = nLcdXOff + (nLcdXPixel * COL_PER_CHAR) + (12-1) * nLcdDistance;
	nLcdYSize = nLcdYOff + (nLcdYPixel * PIX_PER_COL);

	// scan all annunciators
	for (i = 0; i < ARRAYSIZEOF(pAnnunciator); ++i)
	{
		nX = pAnnunciator[i].nOx + pAnnunciator[i].nCx;
		if (nX > nLcdXSize) nLcdXSize = nX;	// new X-max value?

		nY = pAnnunciator[i].nOy + pAnnunciator[i].nCy;
		if (nY > nLcdYSize) nLcdYSize = nY;	// new Y-max value?
	}
	return;
}

//
// draw one display column
//
static __inline void DrawColumn(UINT x, UINT y, INT nNum)
{
	LPBYTE pbyDot = &pbyMask[y * nLcdWidth + x];
	INT i;

	nNum = ~nNum;							// put inverse bit into mask

	for (i = 0; i < 7; ++i, nNum >>= 1)		// draw 7 bit
	{
		BYTE byVal = nNum & 1;

		for (y = 0; y < nLcdYPixel * nAppZoom; ++y) // draw each line
		{
			for (x = 0; x < nLcdXPixel * nAppZoom; ++x)
			{
				pbyDot[x] = byVal;
			}
			pbyDot += nLcdWidth;			// next line
		}
	}
	return;
}

//
// select bitmap object with zoom
//
static HBITMAP SelectObjectZoom(HDC hdc,HBITMAP hgdibmp)
{
	if (nAppZoom > 1)						// adjust background image size to target size
	{
		HDC     hOrgDC;
		HBITMAP hOrgBitmap;
		BITMAP  bm;

		hOrgBitmap = hgdibmp;				// original bitmap

		VERIFY(hOrgDC = CreateCompatibleDC(hWindowDC));
		GetObject(hOrgBitmap,sizeof(bm),&bm); // get bitmap dimensions
		VERIFY(hOrgBitmap = SelectObject(hOrgDC,hOrgBitmap));
		_ASSERT(hOrgBitmap != (HANDLE) GDI_ERROR);

		VERIFY(hgdibmp = CreateCompatibleBitmap(hWindowDC,bm.bmWidth * nAppZoom,bm.bmHeight * nAppZoom));
		hgdibmp = SelectObject(hdc,hgdibmp);
		_ASSERT(hgdibmp != (HANDLE) GDI_ERROR);

		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			// fill bitmap with nAppZoom factor
			StretchBlt(hdc,
					   0, 0, bm.bmWidth * nAppZoom, bm.bmHeight * nAppZoom,
					   hOrgDC,
					   0, 0, bm.bmWidth, bm.bmHeight,
					   SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
	
		DeleteObject(SelectObject(hOrgDC,hOrgBitmap));
		DeleteDC(hOrgDC);
	}
	else
	{
		hgdibmp = SelectObject(hdc,hgdibmp);
		_ASSERT(hgdibmp != (HANDLE) GDI_ERROR);
	}
	return hgdibmp;
}

//
// create background bitmap with segments
//
static VOID SetBkBitmap(VOID)
{
	_ASSERT(hMainDC);						// check for background
	_ASSERT(hBmpBkDC);						// check for LCD background
	_ASSERT(hSelAnnunDC);					// check for annunciator symbols

	EnterCriticalSection(&csGDILock);		// solving NT GDI problems
	{
		// get original background from bitmap
		BitBlt(hBmpBkDC,
			   0, 0, nLcdXSize * nAppZoom, nLcdYSize * nAppZoom,
			   hMainDC,
			   nLcdX * nAppZoom, nLcdY * nAppZoom,
			   SRCCOPY);

		// display on and background contrast defined ?
		if ((Chipset.IORam[CTRL]&DON) && dwKMLColor[Chipset.IORam[CONTRAST]+16] != I)
		{
			DWORD d;

			HBRUSH hBrush = CreateSolidBrush(RGBtoCOLORREF(dwKMLColor[Chipset.IORam[CONTRAST]+16]));
			hBrush = SelectObject(hBmpBkDC,hBrush);

			// set mask bitmap to background
			FillMemory(pbyMask,nLcdWidth * nLcdYSize * nAppZoom,1);

			// scan complete display area
			for (d = PDISP_BEGIN; d <= PDISP_END; d += 2)
			{
				INT nByteNo = d >> 1;

				// column pixel
				DrawColumn((((nByteNo) / COL_PER_CHAR) * nLcdDistance + ((nByteNo) % COL_PER_CHAR) * nLcdXPixel + nLcdXOff) * nAppZoom,
						   nLcdYOff * nAppZoom,
						   0x7F);

				// annunciators
				if (pAnnunciator[nByteNo].nCx > 0)
				{
					BitBlt(hMaskDC,
						   pAnnunciator[nByteNo].nOx * nAppZoom, pAnnunciator[nByteNo].nOy * nAppZoom,
						   pAnnunciator[nByteNo].nCx * nAppZoom, pAnnunciator[nByteNo].nCy * nAppZoom,
						   hSelAnnunDC,
						   pAnnunciator[nByteNo].nDx * nAppZoom, pAnnunciator[nByteNo].nDy * nAppZoom,
						   SRCAND);
				}
			}

			// mask segment mask with background and brush
			BitBlt(hBmpBkDC, 0, 0, nLcdXSize * nAppZoom, nLcdYSize * nAppZoom, hMaskDC, 0, 0, ROP_PSDPxax);

			DeleteObject(SelectObject(hBmpBkDC,hBrush));
		}
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

static VOID UpdateContrastSaca(VOID)
{
	_ASSERT(hLcdDC != NULL);
	EnterCriticalSection(&csGDILock);		// asynchronous display update needs brush
	{
		// replace old with new brush
		DeleteObject(SelectObject(hLcdDC,CreateSolidBrush(
			RGBtoCOLORREF(dwKMLColor[Chipset.IORam[CONTRAST]]))));

		SetBkBitmap();						// update background bitmap
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

static VOID UpdateContrastLewis(VOID)
{
	BYTE byContrast = 0;					// contrast value for display off

	_ASSERT(nCurrentHardware == HDW_LEWIS);

	// when display is on fetch contrast value
	if ((Chipset.IORam[DSPCTL] & DON) != 0)
	{
		byContrast = ((Chipset.IORam[DSPCTL] & CONT4) << 3) | Chipset.IORam[CONTRAST];
	}

	bmiLcd.bmiColors[1] = *(RGBQUAD*)&dwKMLColor[byContrast];    // pixel on  color
	bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[byContrast+32]; // pixel off color

	// if background color is undefined, use color 0 for compatibility
	if (I == *(DWORD*)&bmiLcd.bmiColors[0]) bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[0];

	// update palette information
	EnterCriticalSection(&csGDILock);		// asynchronous display update!
	{
		_ASSERT(hLcdDC);
		DeleteObject(SelectObject(hLcdDC,CreateDIBSection(hWindowDC,
			(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,&pbyLcd,NULL,0)));
		ZeroMemory(pbyLcd,bmiLcd.Lcd_bmih.biWidth * nLcdYSize * nAppZoom);
	}
	LeaveCriticalSection(&csGDILock);
	UpdateAnnunciators();					// adjust annunciator color
	return;
}

VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	dwKMLColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
	return;
}

VOID GetSizeLcdBitmap(INT *pnX, INT *pnY, INT *pnRot180)
{
	_ASSERT(hLcdDC != NULL);				// following variables are valid
	*pnX = nLcdXSize;
	*pnY = nLcdYSize;
	*pnRot180 = nRot180;
	return;
}

VOID CreateLcdBitmap(VOID)
{
	INT i;

	_ASSERT(nCurrentHardware != HDW_UNKNOWN);

	if (nCurrentHardware == HDW_LEWIS)
	{
		INT nLcdBmpWidth,nLcdHeight;

		_ASSERT(nCurrentHardware == HDW_LEWIS);

		// create LCD bitmap
		if (nVertical > 0)					// landscape mode
		{
			if (cCurrentRomType != 'O' && cCurrentRomType != 'Y')
			{
				// set display update routine
				UpdateDisplayMem = UpdateDisplayPioneerV;

				nLcdBmpWidth = LCD_ROW_V16;
				nLcdWidth	 = 16;
				nLcdHeight   = 131;

				if (nLcdyZoom == (UINT) -1)	// not initialized
				{
					switch (nLcdZoom)
					{
						default:
						case 1 : nLcdxZoom = 2; break;
						case 2 : nLcdxZoom = 3; break;
						case 3 : nLcdxZoom = 4; break;
						case 4 : nLcdxZoom = 6; break;
					}
				}
				else						// initialized
				{
					nLcdxZoom = nLcdyZoom;
				}
			}
			else
			{
				// set display update routine
				UpdateDisplayMem = UpdateDisplayClamshellV;

				nLcdBmpWidth = LCD_ROW_V32;
				nLcdWidth	 = 32;
				nLcdHeight   = 137;

				if (nLcdyZoom == (UINT) -1)	// not initialized
				{
					nLcdxZoom = nLcdZoom;
				}
				else						// initialized
				{
					nLcdxZoom = nLcdyZoom;
				}
			}

			nLcdyZoom = nLcdZoom;			// preset y zoom variable

			// rotating 180 degree flag
			nRot180 = (nVertical > 1) ? -1 : 1;
		}
		else								// portrait mode
		{
			nLcdBmpWidth = LCD_ROW_H;

			nLcdxZoom = nLcdZoom;			// preset x zoom variable

			if (cCurrentRomType != 'O' && cCurrentRomType != 'Y')
			{
				// set display update routine
				UpdateDisplayMem = UpdateDisplayPioneerH;

				nLcdWidth  = 131;
				nLcdHeight = 16;

				if (nLcdyZoom == (UINT) -1)	// not initialized
				{
					switch (nLcdxZoom)
					{
						default:
						case 1 : nLcdyZoom = 2; break;
						case 2 : nLcdyZoom = 3; break;
						case 3 : nLcdyZoom = 4; break;
						case 4 : nLcdyZoom = 6; break;
					}
				}
			}
			else
			{
				// set display update routine
				UpdateDisplayMem = UpdateDisplayClamshellH;

				nLcdWidth  = 137;
				nLcdHeight = 32;

				if (nLcdyZoom == (UINT) -1)	// not initialized
				{
					nLcdyZoom = nLcdZoom;
				}
			}

			nRot180 = 1;					// no rotation
		}

		// preset unused fields of KML contrast palette
		if (dwKMLColor[0] == I) dwKMLColor[0] = W;
		for (i = 1; i < 32; ++i)
		{
			_ASSERT(i < ARRAYSIZEOF(dwKMLColor));
			if (dwKMLColor[i] == I) dwKMLColor[i] = B;
		}

		// display contrast routine
		UpdateContrast = UpdateContrastLewis;

		nLcdXSize = nLcdWidth  * nLcdxZoom;
		nLcdYSize = nLcdHeight * nLcdyZoom;

		bmiLcd.Lcd_bmih.biWidth = nLcdBmpWidth * nLcdxZoom * nAppZoom;
		bmiLcd.Lcd_bmih.biHeight = nLcdYSize * nAppZoom;
		if (nVertical == 0) bmiLcd.Lcd_bmih.biHeight = -bmiLcd.Lcd_bmih.biHeight;
		_ASSERT(hLcdDC == NULL);
		VERIFY(hLcdDC = CreateCompatibleDC(hWindowDC));
		VERIFY(hLcdBitmap = CreateDIBSection(hWindowDC,(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,&pbyLcd,NULL,0));
		// fill with background color at startup
		ZeroMemory(pbyLcd, bmiLcd.Lcd_bmih.biWidth * nLcdYSize);
		hLcdBitmap = SelectObject(hLcdDC,hLcdBitmap);
		_ASSERT(hPalette != NULL);
		SelectPalette(hLcdDC,hPalette,FALSE); // set palette for LCD DC
		RealizePalette(hLcdDC);				// realize palette
		UpdateContrast();					// fill with background color at startup
	}
	else									// Sacajavea
	{
		_ASSERT(nCurrentHardware == HDW_SACA);

		GetDimLcdBitmap();					// get LCD bitmap dimension from annunciators
		nRot180 = 1;						// no rotation

		// preset unused fields of KML contrast palette
		for (i = 0; i < 16; ++i)
		{
			_ASSERT(i < ARRAYSIZEOF(dwKMLColor));
			if (dwKMLColor[i] == I) dwKMLColor[i] = B;
		}

		// set display update routine
		UpdateDisplayMem = UpdateDisplaySacajawea;

		// display contrast routine
		UpdateContrast = UpdateContrastSaca;

		// initialize background bitmap
		VERIFY(hBmpBkDC = CreateCompatibleDC(hWindowDC));
		VERIFY(hBmpBk = CreateCompatibleBitmap(hWindowDC,nLcdXSize * nAppZoom,nLcdYSize * nAppZoom));
		VERIFY(hBmpBk = SelectObject(hBmpBkDC,hBmpBk));
		_ASSERT(hBmpBk != (HANDLE) GDI_ERROR);

		// create mask bitmap
		bmiLcd.Lcd_bmih.biWidth = nLcdXSize * nAppZoom;
		bmiLcd.Lcd_bmih.biHeight = -(LONG) nLcdYSize * nAppZoom;
		*(DWORD *)&bmiLcd.bmiColors[0] = 0x00000000;
		*(DWORD *)&bmiLcd.bmiColors[1] = 0x00FFFFFF;
		VERIFY(hMaskDC = CreateCompatibleDC(hWindowDC));
		VERIFY(hMaskBitmap = CreateDIBSection(hWindowDC,(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,(VOID **)&pbyMask,NULL,0));
		VERIFY(hMaskBitmap = SelectObject(hMaskDC,hMaskBitmap));
		_ASSERT(hMaskBitmap != (HANDLE) GDI_ERROR);
		nLcdWidth = WIDTHBYTES(bmiLcd.Lcd_bmih.biWidth * bmiLcd.Lcd_bmih.biBitCount);

		// create LCD bitmap
		_ASSERT(hLcdDC == NULL);
		VERIFY(hLcdDC = CreateCompatibleDC(hWindowDC));
		VERIFY(hLcdBitmap = CreateCompatibleBitmap(hWindowDC,nLcdXSize * nAppZoom,nLcdYSize * nAppZoom));
		VERIFY(hLcdBitmap = SelectObject(hLcdDC,hLcdBitmap));
		_ASSERT(hLcdBitmap != (HANDLE) GDI_ERROR);

		// set brush for display pattern
		VERIFY(hBrush = SelectObject(hLcdDC,CreateSolidBrush(RGBtoCOLORREF(dwKMLColor[Chipset.IORam[CONTRAST]]))));
		_ASSERT(hBrush != (HANDLE) GDI_ERROR);

		_ASSERT(hPalette != NULL);
		SelectPalette(hLcdDC,hPalette,FALSE);	// set palette for LCD DC
		RealizePalette(hLcdDC);				// realize palette

		UpdateContrast();					// initialize background
		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			_ASSERT(hMainDC);				// background bitmap must be loaded

			// get original background from bitmap
			BitBlt(hLcdDC,
				   0, 0,
				   nLcdXSize * nAppZoom, nLcdYSize * nAppZoom,
				   hMainDC,
				   nLcdX * nAppZoom, nLcdY * nAppZoom,
				   SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
	}
	return;
}

VOID DestroyLcdBitmap(VOID)
{
	WORD i;

	// set contrast palette to startup colors
	for (i = 0; i < ARRAYSIZEOF(dwKMLColor); ++i)
		dwKMLColor[i] = I;

	UpdateContrast = NULL;
	UpdateDisplayMem = NULL;

	if (hLcdDC != NULL)
	{
		nLcdXSize = nLcdYSize = 0;

		if (nCurrentHardware == HDW_SACA)
		{
			// destroy background bitmap
			DeleteObject(SelectObject(hBmpBkDC,hBmpBk));
			DeleteDC(hBmpBkDC);

			// destroy display pattern brush
			DeleteObject(SelectObject(hLcdDC,hBrush));

			// destroy mask bitmap
			DeleteObject(SelectObject(hMaskDC,hMaskBitmap));
			DeleteDC(hMaskDC);
		}

		// destroy LCD bitmap
		DeleteObject(SelectObject(hLcdDC,hLcdBitmap));
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
	}
	return;
}

BOOL CreateMainBitmap(LPCTSTR szFilename)
{
	_ASSERT(hWindowDC != NULL);
	VERIFY(hMainDC = CreateCompatibleDC(hWindowDC));
	if (hMainDC == NULL) return FALSE;		// quit if failed
	_ASSERT(hPalette == NULL);				// resource free
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		hMainDC = NULL;
		return FALSE;
	}
	// select object with nAppZoom zoom factor
	hMainBitmap = SelectObjectZoom(hMainDC,hMainBitmap);
	_ASSERT(hMainBitmap != (HANDLE) GDI_ERROR);

	_ASSERT(hPalette != NULL);
	VERIFY(SelectPalette(hMainDC,hPalette,FALSE));
	RealizePalette(hMainDC);

	hSelAnnunDC = hMainDC;					// use background bitmap for annunciators
	return TRUE;
}

VOID DestroyMainBitmap(VOID)
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		DeleteObject(SelectObject(hMainDC,hMainBitmap));
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;

		hSelAnnunDC = NULL;					// reset annunciator DC
	}
	return;
}

//
// load Sacajawea annunciator bitmap
//
BOOL CreateAnnunBitmap(LPCTSTR szFilename)
{
	_ASSERT(hWindowDC != NULL);
	VERIFY(hAnnunDC = CreateCompatibleDC(hWindowDC));
	if (hAnnunDC == NULL) return FALSE;		// quit if failed
	hAnnunBitmap = LoadBitmapFile(szFilename);
	if (hAnnunBitmap == NULL)
	{
		DeleteDC(hAnnunDC);
		hAnnunDC = NULL;
		return FALSE;
	}
	hAnnunBitmap = SelectObject(hAnnunDC,hAnnunBitmap);

	hSelAnnunDC = hAnnunDC;					// use annunciator bitmap for annunciators
	return TRUE;
}

//
// destroy Sacajawea annunciator bitmap
//
VOID DestroyAnnunBitmap(VOID)
{
	if (hAnnunDC != NULL)
	{
		DeleteObject(SelectObject(hAnnunDC,hAnnunBitmap));
		DeleteDC(hAnnunDC);
		hAnnunDC = NULL;
		hAnnunBitmap = NULL;

		hSelAnnunDC = NULL;					// reset annunciator DC
	}
	return;
}

//****************
//*
//* LCD functions
//*
//****************

static VOID WritePixelH(BYTE *p, BYTE a)
{
	UINT y,s;

	if (nLcdxZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		for (y = 0; y < nLcdyZoom; ++y)
		{
			s = y * 4 * LCD_ROW_H;
			*(DWORD *)&p[s] = dwPattern[a&1];

			s += 4 * nLcdyZoom * LCD_ROW_H;
			*(DWORD *)&p[s] = dwPattern[(a>>1)&1];

			s += 4 * nLcdyZoom * LCD_ROW_H;
			*(DWORD *)&p[s] = dwPattern[(a>>2)&1];

			s += 4 * nLcdyZoom * LCD_ROW_H;
			*(DWORD *)&p[s] = dwPattern[(a>>3)&1];
		}
	}
	if (nLcdxZoom == 3)
	{
		for (y = 0; y < nLcdyZoom; ++y)
		{
			s = y * 3 * LCD_ROW_H;
			p[s+0] = p[s+1] = p[s+2] = a&1;

			s += 3 * nLcdyZoom * LCD_ROW_H;
			p[s+0] = p[s+1] = p[s+2] = (a>>1)&1;

			s += 3 * nLcdyZoom * LCD_ROW_H;
			p[s+0] = p[s+1] = p[s+2] = (a>>2)&1;

			s += 3 * nLcdyZoom * LCD_ROW_H;
			p[s+0] = p[s+1] = p[s+2] = (a>>3)&1;
		}
		return;
	}
	if (nLcdxZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101 };

		_ASSERT((DWORD) p % sizeof(WORD) == 0); // check alignment for ARM CPU

		for (y = 0; y < nLcdyZoom; ++y)
		{
			s = y * 2 * LCD_ROW_H;
			*(WORD *)&p[s] = wPattern[a&1];

			s += 2 * nLcdyZoom * LCD_ROW_H;
			*(WORD *)&p[s] = wPattern[(a>>1)&1];

			s += 2 * nLcdyZoom * LCD_ROW_H;
			*(WORD *)&p[s] = wPattern[(a>>2)&1];

			s += 2 * nLcdyZoom * LCD_ROW_H;
			*(WORD *)&p[s] = wPattern[(a>>3)&1];
		}
	}
	if (nLcdxZoom == 1)
	{
		for (y = 0; y < nLcdyZoom; ++y)
		{
			s = y * LCD_ROW_H;
			p[s] = a&1;

			s += nLcdyZoom * LCD_ROW_H;
			p[s] = (a>>1)&1;

			s += nLcdyZoom * LCD_ROW_H;
			p[s] = (a>>2)&1;

			s += nLcdyZoom * LCD_ROW_H;
			p[s] = (a>>3)&1;
		}
	}
	return;
}

static VOID UpdateDisplayClamshellH(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdxZoom == 1 || nLcdxZoom == 2 || nLcdxZoom == 3 || nLcdxZoom == 4);

	// scan complete display area of master
	for (d = MDISP_BEGIN; d <= MDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x7)*4;
		x0 = (SDISP_END-SDISP_BEGIN+1+d)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelH(p,Chipset.IORam[d]);
	}

	// scan complete display area of slave
	for (d = SDISP_BEGIN; d <= SDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = ((d-SDISP_BEGIN)&0x7)*4;
		x0 = (d-SDISP_BEGIN)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelH(p,(BYTE)(ChipsetS.bLcdSyncErr ? 0xF : ChipsetS.IORam[d]));
	}
	return;
}

static VOID UpdateDisplayPioneerH(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdxZoom == 1 || nLcdxZoom == 2 || nLcdxZoom == 3 || nLcdxZoom == 4);

	// scan complete display area
	for (d = DISP_BEGIN; d <= DISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x3)*4;
		x0 = 66*((d>>2)&1)+(d>>3);

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelH(p,Chipset.IORam[d]);
	}
	return;
}

static VOID WritePixelV(BYTE *p, BYTE a)
{
	LPBYTE d;
	UINT   i;

	// draw 1st line
	for (i = 0; i < nLcdxZoom; ++i)
	{
		d = &p[i];	    *d = (a>>0)&1;
		d += nLcdxZoom; *d = (a>>1)&1;
		d += nLcdxZoom; *d = (a>>2)&1;
		d += nLcdxZoom; *d = (a>>3)&1;
	}

	// copy lines
	for (d = p, i = 1; i < nLcdyZoom; ++i)
	{
		d += bmiLcd.Lcd_bmih.biWidth / nAppZoom;
		memcpy(d,p,nLcdxZoom * 4);
	}
	return;
}

static VOID UpdateDisplayClamshellV(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdyZoom == 1 || nLcdyZoom == 2 || nLcdyZoom == 3 || nLcdyZoom == 4);

	// scan complete display area of master
	for (d = MDISP_BEGIN; d <= MDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x7)*4;
		x0 = (SDISP_END-SDISP_BEGIN+1+d)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V32*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelV(p,Chipset.IORam[d]);
	}

	// scan complete display area of slave
	for (d = SDISP_BEGIN; d <= SDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = ((d-SDISP_BEGIN)&0x7)*4;
		x0 = (d-SDISP_BEGIN)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V32*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelV(p,(BYTE)(ChipsetS.bLcdSyncErr ? 0xF : ChipsetS.IORam[d]));
	}
	return;
}

static VOID UpdateDisplayPioneerV(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdyZoom == 1 || nLcdyZoom == 2 || nLcdyZoom == 3 || nLcdyZoom == 4);

	// scan complete display area
	for (d = DISP_BEGIN; d <= DISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x3)*4;
		x0 = 66*((d>>2)&1)+(d>>3);

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V16*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelV(p,Chipset.IORam[d]);
	}
	return;
}

static VOID UpdateDisplaySacajawea(VOID)
{
	EnterCriticalSection(&csGDILock);		// solving NT GDI problems
	{
		// set mask bitmap to background
		FillMemory(pbyMask,nLcdWidth * nLcdYSize * nAppZoom,1);

		if ((Chipset.IORam[DSPCTL]&DON))
		{
			DWORD d;

			// scan complete display area
			for (d = PDISP_BEGIN; d <= PDISP_END; d += 2)
			{
				INT nByteNo = d >> 1;

				INT nValue = (Chipset.IORam[d+1] << 4) | Chipset.IORam[d];

				// column pixel
				DrawColumn((((nByteNo) / COL_PER_CHAR) * nLcdDistance + ((nByteNo) % COL_PER_CHAR) * nLcdXPixel + nLcdXOff) * nAppZoom,
						   nLcdYOff * nAppZoom,
						   nValue);

				// annunciator set
				if ((nValue & 0x80) != 0 && pAnnunciator[nByteNo].nCx > 0)
				{
					BitBlt(hMaskDC,
						   pAnnunciator[nByteNo].nOx * nAppZoom, pAnnunciator[nByteNo].nOy * nAppZoom,
						   pAnnunciator[nByteNo].nCx * nAppZoom, pAnnunciator[nByteNo].nCy * nAppZoom,
						   hSelAnnunDC,
						   pAnnunciator[nByteNo].nDx * nAppZoom, pAnnunciator[nByteNo].nDy * nAppZoom,
						   SRCAND);			// areas don't overlap but PPC2002 made sometimes trouble with SRCCOPY
				}
			}
		}

		// load lcd with mask bitmap
		BitBlt(hLcdDC, 0, 0, nLcdXSize * nAppZoom, nLcdYSize * nAppZoom, hMaskDC, 0, 0, SRCCOPY);

		// mask segment mask with background and brush
		BitBlt(hLcdDC, 0, 0, nLcdXSize * nAppZoom, nLcdYSize * nAppZoom, hBmpBkDC, 0, 0, ROP_PDSPxax);
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

static __inline VOID DoubleBitmap(VOID)
{
	CONST DWORD dwPattern[] = { 0x00000000, 0x00000101, 0x01010000, 0x01010101};

	INT x,y;

	// scan each row
	for (y = (nLcdYSize - 1) * 2; y >= 0; y -= 2)
	{
		INT nOffset = y * bmiLcd.Lcd_bmih.biWidth;
		LPDWORD lpdwTarget = (LPDWORD) (pbyLcd + nOffset);
		LPBYTE  lpbySource = pbyLcd + nOffset / 4;

		// target address on DWORD boundary
		_ASSERT(((DWORD) lpdwTarget % sizeof(*lpdwTarget)) == 0);

		// scan each column
		for (x = bmiLcd.Lcd_bmih.biWidth / sizeof(*dwPattern); --x >= 0;)
		{
			// double two columns
			lpdwTarget[x] = dwPattern[lpbySource[x*2+1] * 2 + lpbySource[x*2]];
		}

		// double the line
		memcpy((LPBYTE)lpdwTarget+bmiLcd.Lcd_bmih.biWidth,lpdwTarget,bmiLcd.Lcd_bmih.biWidth);
	}
	return;
}

static DWORD WINAPI LcdThread(LPVOID pParam)
{
	while (WaitForSingleObject(hEventLcd,DISPLAY_FREQ) == WAIT_TIMEOUT)
	{
		EnterCriticalSection(&csGDILock);		// solving NT GDI problems
		{
			_ASSERT(UpdateDisplayMem);			// function not initialized
			UpdateDisplayMem();					// update display memory
			if (nAppZoom == 2 && nCurrentHardware == HDW_LEWIS)
				DoubleBitmap();					// double dot matrix target bitmap

			StretchBlt(hWindowDC,
					   (nLcdX - nBackgroundX) * nAppZoom, (nLcdY - nBackgroundY) * nAppZoom,
					   nRot180 * nLcdXSize * nAppZoom, nRot180 * nLcdYSize * nAppZoom,
					   hLcdDC,
					   0, 0,
					   nLcdXSize * nAppZoom, nLcdYSize * nAppZoom,
					   SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
	}
	return 0;
	UNREFERENCED_PARAMETER(pParam);
}

VOID StartDisplay(VOID)
{
	if (hCThreadLcd)						// LCD update thread running
		return;								// -> quit

	if (Chipset.IORam[DSPCTL]&DON)			// display on?
	{
		// event to cancel Lcd refresh loop
		hEventLcd = CreateEvent(NULL,FALSE,FALSE,NULL);

		_ASSERT(nCurrentHardware != HDW_UNKNOWN);

		if (nCurrentHardware == HDW_LEWIS)
		{
			_ASSERT(nCurrentHardware == HDW_LEWIS);
			UpdateAnnunciators();			// switch on annunciators
		}
		else								// Sacajavea
		{
			_ASSERT(nCurrentHardware == HDW_SACA);
			SetBkBitmap();					// set background
		}

		VERIFY(hCThreadLcd = CreateThread(NULL,0,&LcdThread,NULL,0,NULL));
	}
	return;
}

VOID StopDisplay(VOID)
{
	if (hCThreadLcd == NULL)				// thread stopped
		return;								// -> quit

	SetEvent(hEventLcd);					// leave Lcd update thread
	WaitForSingleObject(hCThreadLcd,INFINITE);
	CloseHandle(hCThreadLcd);
	hCThreadLcd = NULL;						// set flag display update stopped
	CloseHandle(hEventLcd);					// close Lcd event

	_ASSERT(nCurrentHardware != HDW_UNKNOWN);

	if (nCurrentHardware == HDW_LEWIS)
	{
		_ASSERT(nCurrentHardware == HDW_LEWIS);
		ZeroMemory(pbyLcd,bmiLcd.Lcd_bmih.biWidth * nLcdYSize);
	}
	else									// Sacajavea
	{
		_ASSERT(nCurrentHardware == HDW_SACA);
		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			// get original background from bitmap
			BitBlt(hLcdDC, 0, 0, nLcdXSize * nAppZoom, nLcdYSize * nAppZoom,
				   hMainDC, nLcdX * nAppZoom, nLcdY * nAppZoom, SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
	}
	InvalidateRect(hWnd,NULL,FALSE);
	return;
}

//
// Lewis chip annunciator implementation
//
VOID UpdateAnnunciators(VOID)
{
	const DWORD dwAnnAddrP[] = { LA_ALL, LA_UPDOWN, LA_SHIFT, LA_PRINTER, LA_BUSY, LA_BAT, LA_G, LA_RAD };
	const DWORD dwAnnAddrO[] = { SLA_ALL, SLA_HALT, SLA_SHIFT, SLA_ALPHA, SLA_BUSY, SLA_BAT, SLA_RAD, SLA_PRINTER };

	DWORD  const *dwAnnAddr;
	BOOL   bAnnOn;
	LPBYTE pbySource;
	INT    i,j,nCount;

	if (nCurrentHardware != HDW_LEWIS) return;

	bAnnOn = (Chipset.IORam[DSPCTL] & DON) != 0;

	if (Chipset.bSlave)						// has slave controller
	{
		dwAnnAddr = dwAnnAddrO;
		pbySource = ChipsetS.IORam;			// slave controller

		bAnnOn = bAnnOn && (ChipsetS.IORam[DSPCTL] & DON) != 0;
	}
	else
	{
		dwAnnAddr = dwAnnAddrP;
		pbySource = Chipset.IORam;			// master controller
	}

	// check all annuncators
	for (i = 1; i < ARRAYSIZEOF(dwAnnAddrP); ++i)
	{
		nCount = 0;

		if (bAnnOn)							// annunciators on
		{
			DWORD dwAnn = Npack(pbySource+dwAnnAddr[i],8) ^ Npack(pbySource+dwAnnAddr[0],8);

			// count the number of set bits
			for (;dwAnn != 0; dwAnn >>= 1)
			{
				if ((dwAnn & 0x1) != 0) ++nCount;
			}
		}

		// contrast table entry of annunciator
		j = ((Chipset.IORam[DSPCTL] & CONT4) << 3) | Chipset.IORam[CONTRAST];
		j = j + nCount - 16;
		if (j < 0)  j = 0;
		if (j > 31) j = 31;

		DrawAnnunciator(i, nCount > 0, dwKMLColor[j]);
	}
	return;
}

VOID ResizeWindow(VOID)
{
	RECT rc;

	if (hWnd == NULL) return;				// return if window closed

	GetWindowRect(hWnd, &rc);

	// change client position
	rc.top = ((nFullScreen & TOPBAR_OFF) != 0) ? 0 : lrcTop;
	SetWindowPos(hWnd, NULL,
		rc.left,
		rc.top,
		nBackgroundW * nAppZoom,
		nBackgroundH * nAppZoom,
		SWP_NOZORDER);

	GetClientRect(hWnd, &rc);
	if (rc.bottom < (LONG) (nBackgroundH * nAppZoom))
	{
		SetWindowPos(hWnd, NULL, 0, 0,
			nBackgroundW * nAppZoom,
			(2 * nBackgroundH - rc.bottom) * nAppZoom,
			SWP_NOMOVE | SWP_NOZORDER);
	}
	InvalidateRect(hWnd,NULL,TRUE);
	return;
}
