/*
 *   timer.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"								// I/O definitions

#define AUTO_OFF    10						// Time in minutes for 'auto off'

// Ticks for 01.01.1970 00:00:00
#define UNIX_0_TIME	0x0001cf2e8f800000

// Ticks for 'auto off'
#define OFF_TIME	((LONGLONG) (AUTO_OFF * 60) << 13)

// memory address for clock and auto off

// HP17BII, HP42S = 0x50000-0x50019
#define PIONEER_TIME	0x50000				// clock (=NEXTIRQ)

// HP19BII = 0xC4806-C481D
#define TYCOON2_TIME	0xC4806				// clock (=NEXTIRQ)

// HP28S = 0xC0003-0xC000E, 0xC015A
#define ORLANDO_TIME	0xC0003				// clock (=NEXTIRQ)
#define ORLANDO_OFFCNT	0xC015A				// one minute off counter

#define T1_FREQ		62						// Timer1 1/frequency in ms
#define T2_FREQ		8192					// Timer2 frequency

// typecast define for easy access
#define IOReg(c,o) ((sT[c].pIORam)[o])		// independent I/O access

typedef struct
{
	enum CHIP eChip;						// timer chip (MASTER/SLAVE)
	BOOL   bStarted;						// timer state
	HANDLE hCThreadT1;						// thread handle for timer1
	HANDLE hCThreadT2;						// thread handle for timer2
	BOOL   bOutRange;						// flag if timer value out of range
	BYTE   *pT1;							// address to timer1 register
	DWORD  *pT2;							// address to timer2 register
	LPBYTE pIORam;							// address to IO RAM
	
	// begin of uninitialized variables
	HANDLE hEventT1;						// event handle to terminate timer1 thread
	HANDLE hEventT2;						// event handle to terminate timer2 thread
	BOOL   bT2Running;						// thread running flag for timer2
	LARGE_INTEGER lT2Ref;					// counter value at timer2 start
	DWORD dwT2Ref;							// timer2 value at last timer2 access
	DWORD dwT2Cyc;							// cpu cycle counter at last timer2 access
} TIMERDAT, FAR *LPTIMERDAT, *PTIMERDAT;

static TIMERDAT sT[] =						// timer control
{
	{ MASTER, FALSE, NULL, NULL, FALSE, &Chipset.t1,  &Chipset.t2,  Chipset.IORam  },
	{ SLAVE,  FALSE, NULL, NULL, FALSE, &ChipsetS.t1, &ChipsetS.t2, ChipsetS.IORam }
};

static __inline VOID SetT2Refpoint(LPTIMERDAT pT)
{
	pT->dwT2Ref = *(pT->pT2);				// timer2 value at last timer2 access
	pT->dwT2Cyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF); // cpu cycle counter at last timer2 access
	QueryPerformanceCounter(&(pT->lT2Ref));	// time of corresponding Chipset.t2 value
}

static DWORD CalcT2(enum CHIP eChip)		// calculate timer2 value
{
	DWORD dwT2 = *sT[eChip].pT2;			// get value from chipset

	if (sT[eChip].bStarted)					// timer2 running
	{
		LARGE_INTEGER lT2Act;
		DWORD         dwT2Dif;

		// timer should run a little bit faster (10%) than maschine in authentic speed mode
		DWORD dwCycPerTick = (3 * T2CYCLES) / 5;

		QueryPerformanceCounter(&lT2Act);	// actual time
		// calculate realtime timer2 ticks since reference point
		dwT2 -= (DWORD)
				(((lT2Act.QuadPart - sT[eChip].lT2Ref.QuadPart) * T2_FREQ)
				/ lFreq.QuadPart);

		dwT2Dif = sT[eChip].dwT2Ref - dwT2;	// timer2 ticks since last request

		// 2nd timer call in a 156ms time frame or elapsed time is negative (Win2k bug)
		if (!Chipset.Shutdn && ((dwT2Dif > 0x01 && dwT2Dif <= 0x500) || (dwT2Dif & 0x80000000) != 0))
		{
			DWORD dwT2Ticks = ((DWORD) (Chipset.cycles & 0xFFFFFFFF) - sT[eChip].dwT2Cyc) / dwCycPerTick;

			// estimated < real elapsed timer2 ticks or negative time
			if (dwT2Ticks < dwT2Dif || (dwT2Dif & 0x80000000) != 0)
			{
				// real time too long or got negative time elapsed
				dwT2 = sT[eChip].dwT2Ref - dwT2Ticks; // estimated timer2 value from CPU cycles
				sT[eChip].dwT2Cyc += dwT2Ticks * dwCycPerTick; // estimated CPU cycles for the timer2 ticks
			}
			else
			{
				// reached actual time -> new synchronizing
				sT[eChip].dwT2Cyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF) - dwCycPerTick;
			}
		}
		else
		{
			// valid actual time -> new synchronizing
			sT[eChip].dwT2Cyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF) - dwCycPerTick;
		}

		// check if timer2 interrupt is active -> no timer2 value below 0xFFFFFFFF
		if (   Chipset.inte
			&& (dwT2 & 0x80000000) != 0
			&& (!Chipset.Shutdn || (IOReg(eChip,TIMERCTL_2)&WKE))
			&& (IOReg(eChip,TIMERCTL_2)&INTR)
		   )
		{
			dwT2 = 0xFFFFFFFF;
			sT[eChip].dwT2Cyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF) - dwCycPerTick;
		}

		sT[eChip].dwT2Ref = dwT2;			// new reference time
	}
	return dwT2;
}

static VOID CheckT1(enum CHIP eChip, BYTE nT1)
{
	if ((nT1&8) == 0)						// timer1 MSB not set
	{
		IOReg(eChip,TIMERCTL_1) &= ~SRQ;	// clear SRQ bit
		return;
	}

	_ASSERT((nT1&8) != 0);					// timer1 MSB set

	// timer MSB is one and either INT or WAKE is set
	if ((IOReg(eChip,TIMERCTL_1)&(WKE|INTR)) != 0)
		IOReg(eChip,TIMERCTL_1) |= SRQ;		// set SRQ
	// cpu not sleeping and T1 -> Interrupt
	if (   (!Chipset.Shutdn || (IOReg(eChip,TIMERCTL_1)&WKE))
		&& (IOReg(eChip,TIMERCTL_1)&INTR))
	{
		Chipset.SoftInt = TRUE;
		bInterrupt = TRUE;
	}
	// cpu sleeping and T1 -> Wake Up
	if (Chipset.Shutdn && (IOReg(eChip,TIMERCTL_1)&WKE))
	{
		IOReg(eChip,TIMERCTL_1) &= ~WKE;	// clear WKE bit
		Chipset.bShutdnWake = TRUE;			// wake up from SHUTDN mode
		SetEvent(hEventShutdn);				// wake up emulation thread
	}
	return;
}

static VOID CheckT2(enum CHIP eChip, DWORD dwT2)
{
	if ((dwT2&0x80000000) == 0)				// timer2 MSB not set
	{
		IOReg(eChip,TIMERCTL_2) &= ~SRQ;	// clear SRQ bit
		return;
	}

	_ASSERT((dwT2&0x80000000) != 0);		// timer2 MSB set

	// timer MSB is one and either INT or WAKE is set
	if ((IOReg(eChip,TIMERCTL_2)&(WKE|INTR)) != 0)
		IOReg(eChip,TIMERCTL_2) |= SRQ;		// set SRQ
	// cpu not sleeping and T2 -> Interrupt
	if (   (!Chipset.Shutdn || (IOReg(eChip,TIMERCTL_2)&WKE))
		&& (IOReg(eChip,TIMERCTL_2)&INTR))
	{
		Chipset.SoftInt = TRUE;
		bInterrupt = TRUE;
	}
	// cpu sleeping and T2 -> Wake Up
	if (Chipset.Shutdn && (IOReg(eChip,TIMERCTL_2)&WKE))
	{
		IOReg(eChip,TIMERCTL_2) &= ~WKE;	// clear WKE bit
		Chipset.bShutdnWake = TRUE;			// wake up from SHUTDN mode
		SetEvent(hEventShutdn);				// wake up emulation thread
	}
	return;
}

static DWORD WINAPI T1Thread(LPTIMERDAT pT)
{
	while (WaitForSingleObject(pT->hEventT1,T1_FREQ) == WAIT_TIMEOUT)
	{
		EnterCriticalSection(&csT1Lock);
		{
			// decrement timer value
			*(pT->pT1) = (*(pT->pT1) - 1) & 0xF;
			// test timer1 control bits
			CheckT1(pT->eChip,*(pT->pT1));
		}
		LeaveCriticalSection(&csT1Lock);
	}
	return 0;
}

static DWORD WINAPI T2Thread(LPTIMERDAT pT)
{
	DWORD dwDelay;
	BOOL  bOutRange = FALSE;

	while (pT->bT2Running)
	{
		EnterCriticalSection(&csT2Lock);
		{
			if (!bOutRange)					// save reference time
			{
				SetT2Refpoint(pT);			// reference point for CalcT2()
				dwDelay = *pT->pT2;			// timer value for delay
			}
			else							// called without new refpoint, restart t2 with actual value
			{
				dwDelay = CalcT2(pT->eChip); // actual timer value for delay
			}

			// delay too big for multiplication
			if ((bOutRange = dwDelay > (0xFFFFFFFF - 1023) / 125))
				dwDelay = (0xFFFFFFFF - 1023) / 125; // wait maximum delay time
		}
		LeaveCriticalSection(&csT2Lock);

		dwDelay = (dwDelay * 125 + 1023) / 1024; // timer delay in ms (1000/8192 = 125/1024)
		dwDelay = __max(1,dwDelay);			// wait minimum delay of timer

		if (WaitForSingleObject(pT->hEventT2,dwDelay) != WAIT_TIMEOUT)
		{
			// got new timer2 value
			bOutRange = FALSE;				// set new refpoint
			continue;
		}

		EnterCriticalSection(&csT2Lock);
		{
			// timer2 overrun, test timer2 control bits
			*pT->pT2 = CalcT2(pT->eChip);	// calculate new timer2 value
			CheckT2(pT->eChip,*pT->pT2);	// test timer2 control bits
		}
		LeaveCriticalSection(&csT2Lock);
	}
	return 0;
}

static VOID AbortT1(enum CHIP eChip)
{
	_ASSERT(sT[eChip].hCThreadT1);
	SetEvent(sT[eChip].hEventT1);			// leave timer1 update thread
	WaitForSingleObject(sT[eChip].hCThreadT1,INFINITE);
	CloseHandle(sT[eChip].hCThreadT1);
	sT[eChip].hCThreadT1 = NULL;			// set flag display update stopped
	return;
}

static VOID AbortT2(enum CHIP eChip)
{
	_ASSERT(sT[eChip].hCThreadT2);
	sT[eChip].bT2Running = FALSE;			// leave working loop
	SetEvent(sT[eChip].hEventT2);			// leave timer2 update thread
	WaitForSingleObject(sT[eChip].hCThreadT2,INFINITE);
	CloseHandle(sT[eChip].hCThreadT2);
	sT[eChip].hCThreadT2 = NULL;			// set flag display update stopped
	return;
}

VOID SetHPTime(VOID)						// set date and time
{
	SYSTEMTIME ts;
	LONGLONG   ticks, time;
	BYTE       byTicks[sizeof(ticks)*2];	// unpacked ticks buffer
	DWORD      dw;
	WORD       i;

	_ASSERT(sizeof(LONGLONG) == 8);			// check size of datatype

	GetLocalTime(&ts);						// local time, _ftime() cause memory/resource leaks

	// calculate days until 01.01.1970
	dw = (DWORD) ts.wMonth;
	if (dw > 2)
		dw -= 3L;
	else
	{
		dw += 9L;
		--ts.wYear;
	}
	dw = (DWORD) ts.wDay + (153L * dw + 2L) / 5L;
	dw += (146097L * (((DWORD) ts.wYear) / 100L)) / 4L;
	dw +=   (1461L * (((DWORD) ts.wYear) % 100L)) / 4L;
	dw -= 719469L;

	// convert into seconds and add time
	dw = dw * 24L + (DWORD) ts.wHour;
	dw = dw * 60L + (DWORD) ts.wMinute;
	dw = dw * 60L + (DWORD) ts.wSecond;

	// create timerticks = (s + ms) * 8192
	ticks = ((LONGLONG) dw << 13) | (((LONGLONG) ts.wMilliseconds << 10) / 125);

	ticks += UNIX_0_TIME;					// add offset ticks from year 0
	ticks += (LONG) Chipset.t2;				// add actual timer2 value

	time = ticks;							// save for calc. timeout

	// unpack timer ticks
	for (i = 0; i < ARRAYSIZEOF(byTicks); ++i)
	{
		byTicks[i] = (BYTE) ticks & 0xf;
		ticks >>= 4;
	}

	if (Chipset.type == 'Y')				// Tycoon II, HP19BII
	{
		Nwrite(byTicks, TYCOON2_TIME, 13);	// write date and time

		// disable auto off (1 nib minute up counter + upper 10 nibs of 1min time
		ZeroMemory(byTicks,11);
		Nwrite(byTicks, TYCOON2_TIME+13, 11);
		return;
	}

	if (Chipset.type == 'O')				// Orlando, HP28S
	{
		Nwrite(byTicks, ORLANDO_TIME, 12);	// write date and time

		// HP address for timeout 1 minute counter
		byTicks[0] = AUTO_OFF;
		Nwrite(byTicks, ORLANDO_OFFCNT, 1);
		return;
	}

	// default section for Pioneer series
	Nwrite(byTicks, PIONEER_TIME, 13);		// write date and time

	time += OFF_TIME;						// add 10 min for auto off
	for (i = 0; i < 13; ++i)				// unpack timeout time
	{
		byTicks[i] = (BYTE) time & 0xf;
		time >>= 4;
	}

	// HP address for timeout (=TIMEOUT)
	Nwrite(byTicks, PIONEER_TIME+13, 13);	// write time for auto off
	return;
}

VOID StartTimers(enum CHIP eChip)
{
	// try to start slave timer in non existing slave controller or master timer stopped
	if (eChip == SLAVE && (!Chipset.bSlave || (Chipset.IORam[TIMERCTL_2]&RUN) == 0))
		return;

	if (IOReg(eChip,TIMERCTL_2)&RUN)		// start timer1 and timer2 ?
	{
		if (sT[eChip].bStarted)				// timer running
			return;							// -> quit

		sT[eChip].bStarted = TRUE;			// flag timer running
		CheckT1(eChip,*sT[eChip].pT1);		// check for timer1 interrupts
		CheckT2(eChip,*sT[eChip].pT2);		// check for timer2 interrupts

		// event to cancel timer1 loop
		sT[eChip].hEventT1 = CreateEvent(NULL,FALSE,FALSE,NULL);
		sT[eChip].hEventT2= CreateEvent(NULL,FALSE,FALSE,NULL);

		sT[eChip].bT2Running = TRUE;		// don't leave worker thread

		// start timer1 update thread
		VERIFY(sT[eChip].hCThreadT1 = CreateThread(NULL,0,&T1Thread,&sT[eChip],CREATE_SUSPENDED,NULL));
		VERIFY(SetThreadPriority(sT[eChip].hCThreadT1,THREAD_PRIORITY_HIGHEST));
		ResumeThread(sT[eChip].hCThreadT1);

		// start timer2 update thread
		VERIFY(sT[eChip].hCThreadT2 = CreateThread(NULL,0,&T2Thread,&sT[eChip],CREATE_SUSPENDED,NULL));
		VERIFY(SetThreadPriority(sT[eChip].hCThreadT2,THREAD_PRIORITY_HIGHEST));
		SetT2Refpoint(&sT[eChip]);			// reference point for CalcT2()
		ResumeThread(sT[eChip].hCThreadT2);
	}
	return;
}

VOID StopTimers(enum CHIP eChip)
{
	if (eChip == MASTER && sT[SLAVE].bStarted)	// want to stop master timer but slave timer is running
		StopTimers(SLAVE);					// stop slave timer first

	if (!sT[eChip].bStarted)				// timer stopped
		return;								// -> quit
	if (sT[eChip].hCThreadT1 != NULL)		// timer1 running
	{
		// Critical Section handler may cause a dead lock
		AbortT1(eChip);						// stop timer1
	}
	if (sT[eChip].hCThreadT2 != NULL)		// timer2 running
	{
		EnterCriticalSection(&csT2Lock);
		{
			*sT[eChip].pT2 = CalcT2(eChip);	// update chipset timer2 value
		}
		LeaveCriticalSection(&csT2Lock);
		AbortT2(eChip);						// stop outside critical section
	}
	CloseHandle(sT[eChip].hEventT1);		// close timer1 event
	CloseHandle(sT[eChip].hEventT2);		// close timer2 event
	sT[eChip].bStarted = FALSE;
	return;
}

DWORD ReadT2(enum CHIP eChip)
{
	DWORD dwT2;
	EnterCriticalSection(&csT2Lock);
	{
		dwT2 = CalcT2(eChip);				// calculate timer2 value or if stopped last timer value
		CheckT2(eChip,dwT2);				// update timer2 control bits
	}
	LeaveCriticalSection(&csT2Lock);
	return dwT2;
}

VOID SetT2(enum CHIP eChip, DWORD dwValue)
{
	EnterCriticalSection(&csT2Lock);
	{
		*sT[eChip].pT2 = dwValue;			// set new value
		CheckT2(eChip,*sT[eChip].pT2);		// test timer2 control bits
		if (sT[eChip].bStarted)				// timer running
		{
			SetT2Refpoint(&sT[eChip]);		// reference point for CalcT2()
			SetEvent(sT[eChip].hEventT2);	// new delay
		}
	}
	LeaveCriticalSection(&csT2Lock);
	return;
}

BYTE ReadT1(enum CHIP eChip)
{
	BYTE nT1;
	EnterCriticalSection(&csT1Lock);
	{
		nT1 = *sT[eChip].pT1;				// read timer1 value
		CheckT1(eChip,nT1);					// update timer1 control bits
	}
	LeaveCriticalSection(&csT1Lock);
	return nT1;
}

VOID SetT1(enum CHIP eChip, BYTE byValue)
{
	BOOL bEqual;

	_ASSERT(byValue < 0x10);				// timer1 is only a 4bit counter

	EnterCriticalSection(&csT1Lock);
	{
		bEqual = (*sT[eChip].pT1 == byValue); // check for same value
	}
	LeaveCriticalSection(&csT1Lock);
	if (bEqual) return;						// same value doesn't restart timer period

	if (sT[eChip].hCThreadT1 != NULL)		// timer1 running
	{
		AbortT1(eChip);						// stop timer1
	}
	EnterCriticalSection(&csT1Lock);
	{
		*sT[eChip].pT1 = byValue;			// set new timer1 value
		CheckT1(eChip,*sT[eChip].pT1);		// test timer1 control bits
	}
	LeaveCriticalSection(&csT1Lock);
	if (sT[eChip].bStarted)					// timer running
	{
		// restart timer1 to get full period of frequency
		VERIFY(sT[eChip].hCThreadT1 = CreateThread(NULL,0,&T1Thread,&sT[eChip],CREATE_SUSPENDED,NULL));
		VERIFY(SetThreadPriority(sT[eChip].hCThreadT1,THREAD_PRIORITY_HIGHEST));
		ResumeThread(sT[eChip].hCThreadT1);
	}
	return;
}
